#include <stdint.h>
#include <stdio.h>

// ********** type stubs **********

namespace std {class type_info;}

typedef void (*unexpected_handler)();
typedef void (*terminate_handler)();

typedef int _Unwind_Reason_Code;
typedef int _Unwind_Action;

typedef void (*_Unwind_Exception_Cleanup_Fn) (_Unwind_Reason_Code, struct _Unwind_Exception *);

typedef uint32_t _Unwind_EHT_Header;


// ********** _Unwind_Exception **********

#if defined(__arm__)

// _Unwind_Exception WITHOUT alignment attribute (as found in libobjc2)
struct _Unwind_Exception_v1
{
	uint64_t exception_class;
	void (*exception_cleanup)(_Unwind_Reason_Code, struct _Unwind_Exception_v1 *);
	struct
	{
		uint32_t reserved1;
		uint32_t reserved2;
		uint32_t reserved3;
		uint32_t reserved4;
		uint32_t reserved5;
	} unwinder_cache;
	struct
	{
		uint32_t sp;
		uint32_t bitpattern[5];
	} barrier_cache;
	struct
	{
		uint32_t bitpattern[4];
	} cleanup_cache;
	struct
	{
		uint32_t fnstart;
		_Unwind_EHT_Header *ehtp;
		uint32_t additional;
		uint32_t reserved1;
	} pr_cache;
	long long int :0;
};

// _Unwind_Exception WITH alignment attribute (as found in recent libcxxrt and libunwind)
struct _Unwind_Exception_v2
{
	uint64_t exception_class;
	void (*exception_cleanup)(_Unwind_Reason_Code, struct _Unwind_Exception_v2 *);
	struct
	{
		uint32_t reserved1;
		uint32_t reserved2;
		uint32_t reserved3;
		uint32_t reserved4;
		uint32_t reserved5;
	} unwinder_cache;
	struct
	{
		uint32_t sp;
		uint32_t bitpattern[5];
	} barrier_cache;
	struct
	{
		uint32_t bitpattern[4];
	} cleanup_cache;
	struct
	{
		uint32_t fnstart;
		_Unwind_EHT_Header *ehtp;
		uint32_t additional;
		uint32_t reserved1;
	} pr_cache;
	long long int :0;
} __attribute__((__aligned__(8)));

#else

// _Unwind_Exception WITHOUT alignment attribute (as found in libobjc2)
struct _Unwind_Exception_v1
{
	uint64_t exception_class;
	_Unwind_Exception_Cleanup_Fn exception_cleanup;
	uintptr_t private_1;
	uintptr_t private_2;
};

// _Unwind_Exception WITH alignment attribute (as found in recent libcxxrt and libunwind)
struct _Unwind_Exception_v2
{
	uint64_t exception_class;
	_Unwind_Exception_Cleanup_Fn exception_cleanup;
	uintptr_t private_1;
	uintptr_t private_2;
#if __SIZEOF_POINTER__ == 4
	uint32_t reserved[3];
#endif
} __attribute__((__aligned__));

#endif // __arm__



// ********** __cxa_exception **********

// without referenceCount, using unaligned _Unwind_Exception_v1
struct __cxa_exception_v1
{
	std::type_info *exceptionType;
	void (*exceptionDestructor) (void *);
	unexpected_handler unexpectedHandler;
	terminate_handler terminateHandler;
	__cxa_exception_v1 *nextException;
	int handlerCount;
#if defined(__arm__) && !defined(__ARM_DWARF_EH__)
	_Unwind_Exception_v1 *nextCleanup;
	int cleanupCount;
#endif
	int handlerSwitchValue;
	const char *actionRecord;
	const char *languageSpecificData;
	void *catchTemp;
	void *adjustedPtr;
	_Unwind_Exception_v1 unwindHeader;
};

// without referenceCount, using aligned _Unwind_Exception_v2
struct __cxa_exception_v2
{
	std::type_info *exceptionType;
	void (*exceptionDestructor) (void *);
	unexpected_handler unexpectedHandler;
	terminate_handler terminateHandler;
	__cxa_exception_v2 *nextException;
	int handlerCount;
#if defined(__arm__) && !defined(__ARM_DWARF_EH__)
	_Unwind_Exception_v2 *nextCleanup;
	int cleanupCount;
#endif
	int handlerSwitchValue;
	const char *actionRecord;
	const char *languageSpecificData;
	void *catchTemp;
	void *adjustedPtr;
	_Unwind_Exception_v2 unwindHeader;
};

// with referenceCount, using unaligned _Unwind_Exception_v1
struct __cxa_exception_v3
{
#if __LP64__
	uintptr_t referenceCount;
#endif
	std::type_info *exceptionType;
	void (*exceptionDestructor) (void *);
	unexpected_handler unexpectedHandler;
	terminate_handler terminateHandler;
	__cxa_exception_v3 *nextException;
	int handlerCount;
#if defined(__arm__) && !defined(__ARM_DWARF_EH__)
	_Unwind_Exception_v1 *nextCleanup;
	int cleanupCount;
#endif
	int handlerSwitchValue;
	const char *actionRecord;
	const char *languageSpecificData;
	void *catchTemp;
	void *adjustedPtr;
#if !__LP64__
	uintptr_t referenceCount;
#endif
	_Unwind_Exception_v1 unwindHeader;
};

// with referenceCount, using aligned _Unwind_Exception_v2
struct __cxa_exception_v4
{
#if __LP64__
	uintptr_t referenceCount;
#endif
	std::type_info *exceptionType;
	void (*exceptionDestructor) (void *);
	unexpected_handler unexpectedHandler;
	terminate_handler terminateHandler;
	__cxa_exception_v4 *nextException;
	int handlerCount;
#if defined(__arm__) && !defined(__ARM_DWARF_EH__)
	_Unwind_Exception_v2 *nextCleanup;
	int cleanupCount;
#endif
	int handlerSwitchValue;
	const char *actionRecord;
	const char *languageSpecificData;
	void *catchTemp;
	void *adjustedPtr;
#if !__LP64__
	uintptr_t referenceCount;
#endif
	_Unwind_Exception_v2 unwindHeader;
};



// ********************

int main()
{
	printf("_Unwind_Exception_v1\tsize: %zu, alignment: %zu\n", sizeof(_Unwind_Exception_v1), alignof(_Unwind_Exception_v1));
	printf("_Unwind_Exception_v2\tsize: %zu, alignment: %zu\n", sizeof(_Unwind_Exception_v2), alignof(_Unwind_Exception_v2));
	
	printf("\n");
	
	printf("__cxa_exception_v1\tsize: %zu, alignment: %zu\n", sizeof(__cxa_exception_v1), alignof(__cxa_exception_v1));
	printf("__cxa_exception_v2\tsize: %zu, alignment: %zu\n", sizeof(__cxa_exception_v2), alignof(__cxa_exception_v2));
	printf("__cxa_exception_v3\tsize: %zu, alignment: %zu\n", sizeof(__cxa_exception_v3), alignof(__cxa_exception_v3));
	printf("__cxa_exception_v4\tsize: %zu, alignment: %zu\n", sizeof(__cxa_exception_v4), alignof(__cxa_exception_v4));
	
	return 0;
}
